#!/usr/local/bin/ruby
#
# RDFWeb Ruby RDF stuff
# danbri@w3.org

# Overview:
# we use the classes 'Graph', 'Node', and 'Statement'
# 
# todo: 
#  - associate Nodes with Graphs DONE
#  - implement ask(template_statement) method (slog through the variations!)
#  - implement ntriples-based i/o INPROGRESS
#  - use method_missing to catch property queries on nodes DONE
#  - figure out how to test this and Perl version at same time /?
#  - document the stuff it doesn't do w.r.t. RDF specs /?
#  - find out about various Ruby features I'm unclear on (see 'todo:' notes)
#  - continue migrating ask() to return graphs not nodes DONE
#  - add Mozilla-based graph API 
#  - fix the worst inefficiencies (eg. all those new graphs / statements)
#  - make basicrdf into a module, http://www.rubycentral.com/book/tut_modules.html 

# Currently:
# Nodes and Graphs
#
# a node has a graph; this relationship may be transient, fleeting etc
# default is for node to use the Graph.default graph.
# we only ever have one node per URI, so this could be a source of
# confusion if the graph is seen as dominating the node's state, since
# nodes may be in several graphs, and graphs may mention nodes that
# appear elsewhere. Is this just a matter of better documentation needed?



##############################################################################
#
class Node
  @@nodes = {}
  @db
  @graph
  attr_accessor :content, :graph
   
  # constructor (make this private? use get* instead)
  def initialize (content)
    srand()
    @content = content
  end
  def inspect
    "#@content "
  end
    
  # Get a Node given its URI, recycling where available
  def Node.getResource( content, graph=Graph.default() )
    return @@nodes[content] if @@nodes[content] 
    node = Node.new(content)
    node.graph=graph
    @@nodes[content]=node
    return node 
  end

  # get a fresh blank node
  # notes: couldn't see how to have Node.getResource() work 
  def Node.getBlank(graph=Graph.default)
    content = '[genid:' + rand(100000000).to_s() # ugly stopgap. todo: fix!
    node = Node.new(content) 
    node.graph=graph
    @@nodes[content]=node
    return node
  end

  def Node.getLiteral(content, graph=Graph.default)
    return @@nodes[content] if @@nodes[content]
    node = Node.new("\""+content)
    node.graph=graph
    @@nodes["\""+content]=node
    return node
  end

  def to_s
    "#@content"
  end

  def to_nt_term
    return Node.wrap(@content)
  end

  # wrap literals in quotes, data in <> blanks in ???
  # for NTriples/N3 export
  # 
  def Node.wrap(data)

    wrap = "#{data}"    if (data =~ /^"/)
    wrap = ":_#{data}"  if (data =~ /^\[/) 
    wrap = "<#{data}>" unless wrap
    return wrap
    # todo: multiline, XML content, charset etc...
  end

  # catch missing method calls, assume they're rdf property queries
  # todo: think about error handling; what would be sane here?
  #
  def method_missing(methid)
    str = methid.id2name
    # print "Missing method ",str,"\n" # for rdf property access
    lhs,rhs = str.split(/_+/,2)
    # print "LHS: '#{lhs}' RHS: '#{rhs}' \n"
    uri = self.graph.ns(lhs)
    # print "URI: '#{uri}'\n"
    propuri = uri+rhs
    # print "Should query node for property of type #{propuri}\n"
    ans = self.graph.ask(Statement.new(self,Node.getResource(propuri),nil))
    #todo: scalar or list?
    return ans.objects
  end
end


#############################################################################
#

class Graph

  @@defaultdb=0 # reference to default database, made on demand in default()
  def initialize( statements )
    @db={}	  # counter(notyet, just flag) for each statement we've met
    @fp={}
    @bp={}  
    @ns={}
    return unless statements
    statements.each { |statement| @db[statement]=1 } 
    statements.each {|statement| tell(statement) } 
  end


  def Graph.default()
    if (@@defaultdb==nil)
      print "Getting default graph\n"
      @@defaultdb=Graph.new([]) ## should put the basic RDF/RDFS stuff here
      return @@defaultdb
    end
  end


  # XML Namespace registration
  #
  def reg_xmlns(uri, prefix)
    # print "registering an xml namespace: #{uri} as #{prefix} \n"
    @ns[prefix]=uri 
    return uri
  end

  def ns(prefix)
    # print "Looking up '#{prefix}' in registry, #{@ns.inspect} \n"
    return @ns[prefix]
  end
    

  # Add a bunch of statements (merge into tell() maybe?)
  # 
  def tell_all ( statements )
    statements.each { |s| tell(s) }
  end

  # tell the graph something (add and index a statement)
  #
  def tell ( statement )

    @db[statement]=1

    # store objects under subject+predicate
    #
    sp_list = @fp["#{statement.subject} | #{statement.predicate}"]
    if (sp_list)
      # puts "Storing object under EXISTING s/p, statement= #{statement}"
      # puts "Inspecting existing sp_list: #{sp_list.inspect} \n"
      sp_list.push(statement.object) #todo: we should order this list (and po)
    end
    if (!sp_list)
      # puts "Storing object under NEW s/p, statement = #{statement}"
      sp_list = [statement.object]
      @fp["#{statement.subject} | #{statement.predicate}"]=sp_list
    end

    # store subjects under predicate+object
    #
    po_list = @bp["#{statement.predicate} | #{statement.object}"]    
    if (po_list)
      # puts "Storing subject under EXISTING p/o, statement= #{statement}"
      po_list.push(statement.subject) 
    end					# todo: lookup else syntax for Ruby
    if (!po_list)
      # puts "Storing subject under NEW p/o, statement= #{statement}"
      po_list = [statement.subject]
      @bp["#{statement.predicate} | #{statement.object}"]=po_list
    end
  end


  def toNtriples()
    out = "\n# Ruby-RDF NTriple serializer $Id: basicrdf.rb,v 1.7 2001/12/12 12:39:44 danbri Exp $ \n#\n\n"
    # forward pointers -- from subject+predicate to object(s)
    @fp.each_key { |key|
      s,junk, p = key.split(/ | /,3)
      stem = "<#{s}> <#{p}>"
      # print "Value: #{@fp[key]} \n"
      @fp[key].each { |value| out += "#{stem} #{value.to_nt_term} .\n" }
    }
    return out
  end


  # return all the blunt ends of the arcs in this graph
  def subjects()
    ans=[]
    @db.each_key{ |statement| ans.push(statement.subject) } 
    return ans
  end

  # return all the sharp ends of the arcs in this graph
  def objects()
    ans=[]
    @db.each_key{ |statement| ans.push(statement.object) }
    return ans
  end

  # return all the label nodes for the arcs in this graph
  def predicates()
    ans=[]
    @db.each_key{ |statement| ans.push(statement.predicate) }
    return ans
  end
    
  def size()
    return @db.length 
  end

  # Graph.ask
  #
  # Basic query method for our RDF graph. passed a (template) statement w/ nils
  # ...and returns a graph (which we can probe with subjects(), predicates() etc
  #
  # Notes: this is all pretty inefficient, creating new graphs all
  # over the shop when we needn't, new statements etc etc.
  # todo: change @fp and @bp to store refs to statements not nodes
  #       ...and think about how we can return sub-graph matches
  # without having to go index them. maybe do indexing on demand?
  # (see also tell()

  def ask(query)
    # puts "Vapourware ask/query method called, template statement: #{query} "

    # ooo: dump all statements in the graph
    if (query.predicate==nil && query.subject==nil && query.object==nil) 
      dump =[]
      @db.each_key() {|k| dump.push(k)}
      return Graph.new(dump) 
    end
   
    # xxx: is this statement in the graph? 
    if (query.predicate && query.subject && query.object)
      if ( @db[query] == nil) 
        # puts "Test failed: statement #{query.inspect} is not in graph\n"
      return Graph.new( [ ] ) 
      end
      if ( @db[query] >0)
        # puts "Test succeeded: statement #{query.inspect} is in graph\n"
        return Graph.new([query]) if (@db[query])
      end
    end

    # xxo
    if (query.predicate && query.subject && query.object==nil)
      # puts "xxo: get value(s) given sp"
      # puts "subject = '#{query.subject}' predicate= '#{query.predicate}' \n"
      # puts "Answer lookup: "
      #old:     return @fp["#{query.subject} | #{query.predicate}"]
      obs = @fp["#{query.subject} | #{query.predicate}"]
      response=[]
      query.subject.graph=self	# Is this wrong?
      query.predicate.graph=self

      if obs
        # suppress duplicates (on node identity not strings...)
        obs.uniq!
        seen={}
        new=[]
        obs.each {|i| new.push(i) if !(seen[i.to_s]); seen[i.to_s]=1 }
        obs = new.uniq
      end

      obs.each { |object| object.graph=self; response.push(Statement.new(query.subject,query.predicate, object)) } if obs
      ans = Graph.new(response)
      return ans
    end
 
    # puts "oxx: get subjects(s) given po"
    ## TODO: THIS IS INEFFICIENT. STORE STATEMENTS IN FP and BP!!!
    if (query.predicate && query.subject==nil && query.object)
      subs = @bp["#{query.predicate} | #{query.object}"]
      response=[] 
      query.predicate.graph=self
      query.object.graph=self

      if subs
        # suppress duplicates (on node identity not strings...)
        subs.uniq!
        seen={}
        new=[]
        subs.each {|i| new.push(i) if !(seen[i.to_s]); seen[i.to_s]=1 }
        subs = new.uniq
      end

      subs.each { |subject| subject.graph=self; response.push(Statement.new(subject,query.predicate,query.object)) } if subs
      ans = Graph.new(response)
      # puts "Returning a graph! details: #{ans.inspect} \n"
      return ans 
    end 
  
  ## more query facilities needed here

  # done: xxx ooo xxo oxx
  # todo: xox oxo oox xoo 

  end

  # nodes that know... (about a graph)
  def getResource(content)
    return Node.getResource(content, self)
  end

  def getBlank()
    return Node.getBlank(self)
  end

  def getLiteral(content)
    return Node.getLiteral(content, self)
  end

end

###########################################################################
#
class Statement
  attr_accessor :predicate, :subject, :object
  def initialize (subject, predicate, object)

    subject   = Node.getResource(subject) if subject.type == String 
    predicate = Node.getResource(predicate) if predicate.type == String
    object = Node.getResource(object) if object.type == String #todo
    #print "In statement init. #{subject} -- #{predicate} -> #{object} \n"

    @subject = subject
    @predicate = predicate
    @object = object
  end
  def inspect
    "<#@subject> <#@predicate> <#@object> "
  end 
  def to_s
  "Statement: #@subject> <#@predicate> <#@object>\n"
  end

end

############################################################################
#    A rather basic NTriples parser, cut down from:
#    http://www.w3.org/2000/10/swap/n-triples2kif.pl
#    http://www.w3.org/TR/rdf-testcases/#ntriples
# seeAlso: 
#     http://www.rubycentral.com/book/tut_stdtypes.html

class Loader

  # clean up a term t and return
  # litOK: flag whether literals acceptable or not
  def Loader.term (t, litOK)
    t.chomp()
    t=t.sub(/^\s*</,'')
    t=t.sub(/>\s*$/,'') 
    # a whole bunch more stuff see perl script above
    # print "Modified t: #{t.inspect} \n" 
    return t
  end

  # call out to an RDF2Ntriples parser
  def Loader.fn2nt (fn)
    print "Running external parser on filename #{fn} \n"
    ##TODO!
    return 
  end

  def Loader.base
    return '../' # how should we deal with need for a homedir?
  end

  # filename, returning ntriples as text
  #
  def Loader.xsltrdf (fn)
    require 'sablot' #exception handling if this is missing?
    xsl = open(Loader.base+'xsltrdf/rdf2nt.xsl'){|f| f.read}  # need path to XSLT file?
    xml =  open(fn){|f| f.read}  
    return Sablot::process_strings(xsl, xml) 
  end

  # pull NTriples from somewhere, return them as a new Graph
  #
  def Loader.rdf2nt (input, data=Graph.new([]))
    # print "NT reader: #{input.inspect}\n"
    gets.each {|l| Loader.parseline(l,data) } # defaults STDIN
    return data
  end

  # URI (just a filename currently), returns a data graph (somehow)
  #
  def Loader.rdf2graph(uri, data = Graph.new([]) )
    nt = Loader.xsltrdf(uri) # we'll use XSLT
			       # could use web service, commandline etc too
    nt.each { |l| Loader.parseline(l,data) } 
    # print "Got graph in NT: #{data.inspect} \n"
    return data
   end

  def Loader.nt2graph(input='stdin', data=Graph.new([]) )
    # default to STDIN; should allow forfiles too
    while (gets)
      Loader.parseline($_,data)
    end
    return data
  end
  
  def Loader.parseline(line,data)  
      $_ = line # need to tidy this up
      return if /^#/		# re IO see 
      return unless /\S/	# http://www.rubycentral.com/book/intro.html
      $_ = $_.sub(/^ */, '')	# http://www.rubycentral.com/book/tut_io.html
      $_ = $_.sub(/\s*\.\s*$/, '')	# 
      chomp
      # print "LOADER: data is now: '#{$_}' \n"
   

      ## TODO: literals with spaces
      parts = $_.split(/\s+/) # other ws? tricky as need to re-assemble
			    # this is the wrong way to parse ntriples
      # puts "LOADER: Summary: #{parts.inspect}\n"
      return unless parts.length>=3 # literals > 3? 
      parts[2] += " "+parts[3..parts.length].join(' ') if(parts.length >3)
      st = Loader.term(parts[0], 0)
      pt = Loader.term(parts[1], 0)
      ot = Loader.term(parts[2], 1)
      # puts "LOADER: output: s:#{st} p:#{pt} o:#{ot} \n"
      s = Statement.new( Node.new(st), Node.new(pt), Node.new(ot) )
      data.tell(s)
  end
end


